@@ -24,8 +24,9 @@
   SOFTWARE.
 */
 
-#include <errno.h>
-#include <string.h>
+#include <cassert>
+#include <cerrno>
+#include <cstring>
 #include <getopt.h>
 
 #include <fstream>
@@ -35,29 +36,23 @@
 #include <algorithm>
 #include <vector>
 #include <string>
+#include <sstream>
+#include <filesystem>
+#include <regex>
+#include <optional>
 
 #include <libheif/heif.h>
 #include <libheif/heif_properties.h>
 #include "libheif/heif_items.h"
 
-#if HAVE_LIBJPEG
-#include "decoder_jpeg.h"
-#endif
+#include "heifio/decoder_jpeg.h"
+#include "heifio/decoder_png.h"
+#include "heifio/decoder_tiff.h"
+#include "heifio/decoder_y4m.h"
 
-#if HAVE_LIBPNG
-#include "decoder_png.h"
-#endif
-
-#if HAVE_LIBTIFF
-#include "decoder_tiff.h"
-#endif
-
-#include "decoder_y4m.h"
-
-#include <assert.h>
 #include "benchmark.h"
-#include "exif.h"
 #include "common.h"
+#include "libheif/heif_experimental.h"
 
 int master_alpha = 1;
 int thumb_alpha = 1;
@@ -66,8 +61,14 @@ int two_colr_boxes = 0;
 int premultiplied_alpha = 0;
 int run_benchmark = 0;
 int metadata_compression = 0;
+int tiled_input_x_y = 0;
 const char* encoderId = nullptr;
 std::string chroma_downsampling;
+int tiled_image_width = 0;
+int tiled_image_height = 0;
+std::string tiling_method = "grid";
+heif_unci_compression unci_compression = heif_unci_compression_brotli;
+int add_pyramid_group = 0;
 
 uint16_t nclx_colour_primaries = 1;
 uint16_t nclx_transfer_characteristic = 13;
@@ -83,9 +84,7 @@ std::string property_pitm_description;
 #endif
 
 #if HAVE_GETTIMEOFDAY
-
 #include <sys/time.h>
-
 struct timeval time_encoding_start;
 struct timeval time_encoding_end;
 #endif
@@ -101,6 +100,10 @@ const int OPTION_USE_JPEG2000_COMPRESSION = 1007;
 const int OPTION_VERBOSE = 1008;
 const int OPTION_USE_HTJ2K_COMPRESSION = 1009;
 const int OPTION_USE_VVC_COMPRESSION = 1010;
+const int OPTION_TILED_IMAGE_WIDTH = 1011;
+const int OPTION_TILED_IMAGE_HEIGHT = 1012;
+const int OPTION_TILING_METHOD = 1013;
+const int OPTION_UNCI_COMPRESSION = 1014;
 
 
 static struct option long_options[] = {
@@ -125,6 +128,7 @@ static struct option long_options[] = {
     {(char* const) "htj2k",                   no_argument,       0,              OPTION_USE_HTJ2K_COMPRESSION},
 #if WITH_UNCOMPRESSED_CODEC
     {(char* const) "uncompressed",                no_argument,       0,                     'U'},
+    {(char* const) "unci-compression-method",     required_argument, nullptr, OPTION_UNCI_COMPRESSION},
 #endif
     {(char* const) "matrix_coefficients",         required_argument, 0,                     OPTION_NCLX_MATRIX_COEFFICIENTS},
     {(char* const) "colour_primaries",            required_argument, 0,                     OPTION_NCLX_COLOUR_PRIMARIES},
@@ -137,6 +141,12 @@ static struct option long_options[] = {
     {(char* const) "enable-metadata-compression", no_argument,       &metadata_compression, 1},
     {(char* const) "pitm-description",            required_argument, 0,                     OPTION_PITM_DESCRIPTION},
     {(char* const) "chroma-downsampling",         required_argument, 0, 'C'},
+    {(char* const) "tiled-input",                 no_argument, 0, 'T'},
+    {(char* const) "tiled-image-width",           required_argument, nullptr, OPTION_TILED_IMAGE_WIDTH},
+    {(char* const) "tiled-image-height",          required_argument, nullptr, OPTION_TILED_IMAGE_HEIGHT},
+    {(char* const) "tiled-input-x-y",             no_argument,       &tiled_input_x_y, 1},
+    {(char* const) "tiling-method",               required_argument, nullptr, OPTION_TILING_METHOD},
+    {(char* const) "add-pyramid-group",           no_argument,       &add_pyramid_group, 1},
     {0, 0,                                                           0,  0},
 };
 
@@ -173,12 +183,12 @@ void show_help(const char* argv0)
             << "      --jpeg2000        encode as JPEG 2000 (experimental)\n"
             << "      --htj2k           encode as High Throughput JPEG 2000 (experimental)\n"
 #if WITH_UNCOMPRESSED_CODEC
-            << "  -U, --uncompressed    encode as uncompressed image (according to ISO 23001-17) (EXPERIMENTAL)\n"
+            << "  -U, --uncompressed             encode as uncompressed image (according to ISO 23001-17) (EXPERIMENTAL)\n"
+            << "      --unci-compression METHOD  choose one of these methods: none, deflate, zlib, brotli.\n"
 #endif
             << "      --list-encoders         list all available encoders for all compression formats\n"
             << "  -e, --encoder ID            select encoder to use (the IDs can be listed with --list-encoders)\n"
             << "      --plugin-directory DIR  load all codec plugins in the directory\n"
-            << "  -E, --even-size   [deprecated] crop images to even width and height (odd sizes are not decoded correctly by some software)\n"
             << "  --matrix_coefficients     nclx profile: color conversion matrix coefficients, default=6 (see h.273)\n"
             << "  --colour_primaries        nclx profile: color primaries (see h.273)\n"
             << "  --transfer_characteristic nclx profile: transfer characteristics (see h.273)\n"
@@ -191,41 +201,21 @@ void show_help(const char* argv0)
             << "  -C,--chroma-downsampling ALGO   force chroma downsampling algorithm (nn = nearest-neighbor / average / sharp-yuv)\n"
             << "                                  (sharp-yuv makes edges look sharper when using YUV420 with bilinear chroma upsampling)\n"
             << "  --benchmark               measure encoding time, PSNR, and output file size\n"
-            << "  --pitm-description TEXT   (experimental) set user description for primary image\n";
-}
-
-
-#if !HAVE_LIBJPEG
-InputImage loadJPEG(const char* filename)
-{
-  std::cerr << "Cannot load JPEG because libjpeg support was not compiled.\n";
-  exit(1);
-
-  return {};
-}
+            << "  --pitm-description TEXT   (experimental) set user description for primary image\n"
+            << "  -T,--tiled-input          input is a set of tile images (only provide one filename with two tile position numbers).\n"
+            << "                            For example, 'tile-01-05.jpg' would be a valid input filename.\n"
+            << "                            You only have to provide the filename of one tile as input, heif-enc will scan the directory\n"
+            << "                            for the other tiles and determine the range of tiles automatically.\n"
+            << "  --tiled-image-width #     override image width of tiled image\n"
+            << "  --tiled-image-height #    override image height of tiled image\n"
+            << "  --tiled-input-x-y         usually, the first number in the input tile filename should be the y position.\n"
+            << "                            With this option, this can be swapped so that the first number is x, the second number y.\n"
+#if ENABLE_EXPERIMENTAL_FEATURES
+            << "  --tiling-method METHOD    choose one of these methods: grid, tili, unci. The default is 'grid'.\n"
+            << "  --add-pyramid-group       when several images are given, put them into a multi-resolution pyramid group.\n"
 #endif
-
-
-#if !HAVE_LIBPNG
-InputImage loadPNG(const char* filename, int output_bit_depth)
-{
-  std::cerr << "Cannot load PNG because libpng support was not compiled.\n";
-  exit(1);
-
-  return {};
+            ;
 }
-#endif
-
-
-#if !HAVE_LIBTIFF
-InputImage loadTIFF(const char* filename)
-{
-  std::cerr << "Cannot load TIFF because libtiff support was not compiled.\n";
-  exit(1);
-
-  return {};
-}
-#endif
 
 
 void list_encoder_parameters(heif_encoder* encoder)
@@ -373,6 +363,9 @@ static const char* get_compression_format_name(heif_compression_format format)
     case heif_compression_AV1:
       return "AV1";
       break;
+    case heif_compression_AVC:
+      return "AVC";
+      break;
     case heif_compression_VVC:
       return "VVC";
       break;
@@ -399,19 +392,18 @@ static const char* get_compression_format_name(heif_compression_format format)
 
 static void show_list_of_all_encoders()
 {
-    for (auto compression_format : {heif_compression_HEVC, heif_compression_AV1, heif_compression_VVC, heif_compression_JPEG, heif_compression_JPEG2000, heif_compression_HTJ2K
-#if WITH_UNCOMPRESSED_CODEC
-, heif_compression_uncompressed
-#endif
+  for (auto compression_format: {heif_compression_AVC, heif_compression_AV1, heif_compression_HEVC,
+                                 heif_compression_JPEG, heif_compression_JPEG2000, heif_compression_HTJ2K,
+                                 heif_compression_uncompressed, heif_compression_VVC
   }) {
 
     switch (compression_format) {
+      case heif_compression_AVC:
+        std::cout << "AVC";
+        break;
       case heif_compression_AV1:
         std::cout << "AVIF";
         break;
-      case heif_compression_VVC:
-        std::cout << "VVIC";
-        break;
       case heif_compression_HEVC:
         std::cout << "HEIC";
         break;
@@ -422,11 +414,14 @@ static void show_list_of_all_encoders()
         std::cout << "JPEG 2000";
         break;
       case heif_compression_HTJ2K:
-        std::cout << "HT-J2K";
+        std::cout << "JPEG 2000 (HT)";
         break;
       case heif_compression_uncompressed:
         std::cout << "Uncompressed";
         break;
+      case heif_compression_VVC:
+        std::cout << "VVIC";
+        break;
       default:
         assert(false);
     }
@@ -491,6 +486,360 @@ std::string suffix_for_compression_format(heif_compression_format format)
 }
 
 
+InputImage load_image(const std::string& input_filename, int output_bit_depth)
+{
+  InputImage input_image;
+
+  // get file type from file name
+
+  std::string suffix;
+  auto suffix_pos = input_filename.find_last_of('.');
+  if (suffix_pos != std::string::npos) {
+    suffix = input_filename.substr(suffix_pos + 1);
+    std::transform(suffix.begin(), suffix.end(), suffix.begin(), ::tolower);
+  }
+
+  enum
+  {
+    PNG, JPEG, Y4M, TIFF
+  } filetype = JPEG;
+  if (suffix == "png") {
+    filetype = PNG;
+  }
+  else if (suffix == "y4m") {
+    filetype = Y4M;
+  }
+  else if (suffix == "tif" || suffix == "tiff") {
+    filetype = TIFF;
+  }
+
+  if (filetype == PNG) {
+    heif_error err = loadPNG(input_filename.c_str(), output_bit_depth, &input_image);
+    if (err.code != heif_error_Ok) {
+      std::cerr << "Can not load TIFF input_image: " << err.message << '\n';
+      exit(1);
+    }
+  }
+  else if (filetype == Y4M) {
+    heif_error err = loadY4M(input_filename.c_str(), &input_image);
+    if (err.code != heif_error_Ok) {
+      std::cerr << "Can not load TIFF input_image: " << err.message << '\n';
+      exit(1);
+    }
+  }
+  else if (filetype == TIFF) {
+    heif_error err = loadTIFF(input_filename.c_str(), &input_image);
+    if (err.code != heif_error_Ok) {
+      std::cerr << "Can not load TIFF input_image: " << err.message << '\n';
+      exit(1);
+    }
+  }
+  else {
+    heif_error err = loadJPEG(input_filename.c_str(), &input_image);
+    if (err.code != heif_error_Ok) {
+      std::cerr << "Can not load JPEG input_image: " << err.message << '\n';
+      exit(1);
+    }
+  }
+
+  return input_image;
+}
+
+
+heif_error create_output_nclx_profile_and_configure_encoder(heif_encoder* encoder,
+                                                            heif_color_profile_nclx** out_nclx,
+                                                            std::shared_ptr<heif_image> input_image,
+                                                            bool lossless)
+{
+  *out_nclx = heif_nclx_color_profile_alloc();
+  if (!*out_nclx) {
+    return {heif_error_Encoding_error, heif_suberror_Unspecified, "Cannot allocate NCLX color profile."};
+  }
+
+  heif_color_profile_nclx* nclx = *out_nclx; // abbreviation;
+
+  if (lossless) {
+      heif_encoder_set_lossless(encoder, true);
+
+      if (heif_image_get_colorspace(input_image.get()) == heif_colorspace_RGB) {
+        nclx->matrix_coefficients = heif_matrix_coefficients_RGB_GBR;
+        nclx->full_range_flag = true;
+
+        heif_error error = heif_encoder_set_parameter(encoder, "chroma", "444");
+        if (error.code) {
+          return error;
+        }
+      }
+      else {
+        heif_color_profile_nclx* input_nclx = nullptr;
+
+        heif_error error = heif_image_get_nclx_color_profile(input_image.get(), &input_nclx);
+        if (error.code == heif_error_Color_profile_does_not_exist) {
+          // NOP, use default NCLX profile
+        }
+        else if (error.code) {
+          std::cerr << "Cannot get input NCLX color profile.\n";
+          return error;
+        }
+        else {
+          nclx->matrix_coefficients = input_nclx->matrix_coefficients;
+          nclx->transfer_characteristics = input_nclx->transfer_characteristics;
+          nclx->color_primaries = input_nclx->color_primaries;
+          nclx->full_range_flag = input_nclx->full_range_flag;
+
+          heif_nclx_color_profile_free(input_nclx);
+          input_nclx = nullptr;
+        }
+
+        assert(!input_nclx);
+
+        // TODO: this assumes that the encoder plugin has a 'chroma' parameter. Currently, they do, but there should be a better way to set this.
+        switch (heif_image_get_chroma_format(input_image.get())) {
+          case heif_chroma_420:
+          case heif_chroma_monochrome:
+            error = heif_encoder_set_parameter(encoder, "chroma", "420");
+            break;
+          case heif_chroma_422:
+            error = heif_encoder_set_parameter(encoder, "chroma", "422");
+            break;
+          case heif_chroma_444:
+            error = heif_encoder_set_parameter(encoder, "chroma", "444");
+            break;
+          default:
+            assert(false);
+            exit(5);
+        }
+
+        if (error.code) {
+          return error;
+        }
+      }
+  }
+
+  if (!lossless) {
+    heif_error error = heif_nclx_color_profile_set_matrix_coefficients(nclx, nclx_matrix_coefficients);
+    if (error.code) {
+      std::cerr << "Invalid matrix coefficients specified.\n";
+      exit(5);
+    }
+    error = heif_nclx_color_profile_set_transfer_characteristics(nclx, nclx_transfer_characteristic);
+    if (error.code) {
+      std::cerr << "Invalid transfer characteristics specified.\n";
+      exit(5);
+    }
+    error = heif_nclx_color_profile_set_color_primaries(nclx, nclx_colour_primaries);
+    if (error.code) {
+      std::cerr << "Invalid color primaries specified.\n";
+      exit(5);
+    }
+    nclx->full_range_flag = (uint8_t) nclx_full_range;
+  }
+
+  return {heif_error_Ok};
+}
+
+
+struct input_tiles_generator
+{
+  uint32_t first_start;
+  uint32_t first_end;
+  uint32_t first_digits;
+  uint32_t second_start;
+  uint32_t second_end;
+  uint32_t second_digits;
+
+  std::filesystem::path directory;
+  std::string prefix;
+  std::string separator;
+  std::string suffix;
+
+  bool first_is_x = false;
+
+  uint32_t nColumns() const { return first_is_x ? (first_end - first_start + 1) : (second_end - second_start + 1); }
+  uint32_t nRows() const { return first_is_x ? (second_end - second_start + 1) : (first_end - first_start + 1); }
+
+  uint32_t nTiles() const { return (first_end - first_start + 1) * (second_end - second_start + 1); }
+
+  std::filesystem::path filename(uint32_t tx, uint32_t ty) const
+  {
+    std::stringstream sstr;
+
+    sstr << prefix << std::setw(first_digits) << std::setfill('0') << (first_is_x ? tx : ty) + first_start;
+    sstr << separator << std::setw(second_digits) << std::setfill('0') << (first_is_x ? ty : tx) + second_start;
+    sstr << suffix;
+
+    std::filesystem::path p = directory / sstr.str();
+    return p;
+  }
+};
+
+std::optional<input_tiles_generator> determine_input_images_tiling(const std::string& filename)
+{
+  std::regex pattern(R"((.*\D)?(\d+)(\D+?)(\d+)(\..+)$)");
+  std::smatch match;
+
+  input_tiles_generator generator;
+
+  if (std::regex_match(filename, match, pattern)) {
+    std::string prefix = match[1];
+
+    auto p = std::filesystem::absolute(std::filesystem::path(prefix));
+    generator.directory = p.parent_path();
+    generator.prefix = p.filename().string(); // TODO: we could also use u8string(), but it is not well supported in C++20
+
+    generator.separator = match[3];
+    generator.suffix = match[5];
+
+    generator.first_start = 9999;
+    generator.first_end = 0;
+    generator.first_digits = 9;
+
+    generator.second_start = 9999;
+    generator.second_end = 0;
+    generator.second_digits = 9;
+  }
+  else {
+    return std::nullopt;
+  }
+
+  std::string patternString = generator.prefix + "(\\d+)" + generator.separator + "(\\d+)" + generator.suffix + "$";
+  pattern = patternString;
+
+  for (const auto& dirEntry : std::filesystem::directory_iterator(generator.directory))
+  {
+    if (dirEntry.is_regular_file()) {
+      std::string s{dirEntry.path().filename().string()};
+
+      if (std::regex_match(s, match, pattern)) {
+        uint32_t first = std::stoi(match[1]);
+        uint32_t second = std::stoi(match[2]);
+
+        generator.first_digits = std::min(generator.first_digits, (uint32_t)match[1].length());
+        generator.second_digits = std::min(generator.second_digits, (uint32_t)match[2].length());
+
+        generator.first_start = std::min(generator.first_start, first);
+        generator.first_end = std::max(generator.first_end, first);
+        generator.second_start = std::min(generator.second_start, second);
+        generator.second_end = std::max(generator.second_end, second);
+      }
+    }
+  }
+
+  return generator;
+}
+
+
+heif_image_handle* encode_tiled(heif_context* ctx, heif_encoder* encoder, heif_encoding_options* options,
+                                int output_bit_depth,
+                                const input_tiles_generator& tile_generator,
+                                const heif_image_tiling& tiling)
+{
+  heif_image_handle* tiled_image = nullptr;
+
+
+  // --- create the main grid image
+
+  if (tiling_method == "grid") {
+    heif_error error = heif_context_add_grid_image(ctx, tiling.image_width, tiling.image_height,
+                                                   tiling.num_columns, tiling.num_rows,
+                                                   options,
+                                                   &tiled_image);
+    if (error.code != 0) {
+      std::cerr << "Could not generate grid image: " << error.message << "\n";
+      return nullptr;
+    }
+  }
+#if ENABLE_EXPERIMENTAL_FEATURES
+  else if (tiling_method == "tili") {
+    heif_tiled_image_parameters tiled_params{};
+    tiled_params.version = 1;
+    tiled_params.image_width = tiling.image_width;
+    tiled_params.image_height = tiling.image_height;
+    tiled_params.tile_width = tiling.tile_width;
+    tiled_params.tile_height = tiling.tile_height;
+    tiled_params.offset_field_length = 32;
+    tiled_params.size_field_length = 24;
+    tiled_params.tiles_are_sequential = 1;
+
+    heif_error error = heif_context_add_tiled_image(ctx, &tiled_params, options, encoder, &tiled_image);
+    if (error.code != 0) {
+      std::cerr << "Could not generate tili image: " << error.message << "\n";
+      return nullptr;
+    }
+  }
+  else if (tiling_method == "unci") {
+    heif_unci_image_parameters params{};
+    params.version = 1;
+    params.image_width = tiling.image_width;
+    params.image_height = tiling.image_height;
+    params.tile_width = tiling.tile_width;
+    params.tile_height = tiling.tile_height;
+    params.compression = unci_compression;
+
+    std::string input_filename = tile_generator.filename(0, 0);
+    InputImage prototype_image = load_image(input_filename, output_bit_depth);
+
+    heif_error error = heif_context_add_unci_image(ctx, &params, options, prototype_image.image.get(), &tiled_image);
+    if (error.code != 0) {
+      std::cerr << "Could not generate unci image: " << error.message << "\n";
+      return nullptr;
+    }
+  }
+#endif
+  else {
+    assert(false);
+    exit(10);
+  }
+
+
+  // --- add all the image tiles
+
+  std::cout << "encoding tiled image, tile size: " << tiling.tile_width << "x" << tiling.tile_height
+            << " image size: " << tiling.image_width << "x" << tiling.image_height << "\n";
+
+  int tile_width = 0, tile_height = 0;
+
+  for (uint32_t ty = 0; ty < tile_generator.nRows(); ty++)
+    for (uint32_t tx = 0; tx < tile_generator.nColumns(); tx++) {
+      std::string input_filename = tile_generator.filename(tx,ty).string();
+
+      InputImage input_image = load_image(input_filename, output_bit_depth);
+
+      if (tile_width == 0) {
+        tile_width = heif_image_get_primary_width(input_image.image.get());
+        tile_height = heif_image_get_primary_height(input_image.image.get());
+
+        if (tile_width <= 0 || tile_height <= 0) {
+          std::cerr << "Could not read input image size correctly\n";
+          return nullptr;
+        }
+      }
+
+      heif_error error;
+      error = heif_image_extend_to_size_fill_with_zero(input_image.image.get(), tile_width, tile_height);
+      if (error.code) {
+        std::cerr << error.message << "\n";
+      }
+
+      std::cout << "encoding tile " << ty+1 << " " << tx+1
+                << " (of " << tile_generator.nRows() << "x" << tile_generator.nColumns() << ")  \r";
+      std::cout.flush();
+
+      error = heif_context_add_image_tile(ctx, tiled_image, tx, ty,
+                                          input_image.image.get(),
+                                          encoder);
+      if (error.code != 0) {
+        std::cerr << "Could not encode HEIF/AVIF file: " << error.message << "\n";
+        return nullptr;
+      }
+    }
+
+  std::cout << "\n";
+
+  return tiled_image;
+}
+
+
 class LibHeifInitializer
 {
 public:
@@ -518,14 +867,14 @@ int main(int argc, char** argv)
   bool force_enc_jpeg = false;
   bool force_enc_jpeg2000 = false;
   bool force_enc_htj2k = false;
-  bool crop_to_even_size = false;
+  bool use_tiling = false;
 
   std::vector<std::string> raw_params;
 
 
   while (true) {
     int option_index = 0;
-    int c = getopt_long(argc, argv, "hq:Lo:vPp:t:b:AEe:C:"
+    int c = getopt_long(argc, argv, "hq:Lo:vPp:t:b:Ae:C:T"
 #if WITH_UNCOMPRESSED_CODEC
         "U"
 #endif
@@ -572,9 +921,6 @@ int main(int argc, char** argv)
         force_enc_uncompressed = true;
         break;
 #endif
-      case 'E':
-        crop_to_even_size = true;
-        break;
       case 'e':
         encoderId = optarg;
         break;
@@ -619,6 +965,43 @@ int main(int argc, char** argv)
         }
         break;
       }
+      case OPTION_TILED_IMAGE_WIDTH:
+        tiled_image_width = (int) strtol(optarg, nullptr, 0);
+        break;
+      case OPTION_TILED_IMAGE_HEIGHT:
+        tiled_image_height = (int) strtol(optarg, nullptr, 0);
+        break;
+      case OPTION_TILING_METHOD:
+        tiling_method = optarg;
+        if (tiling_method != "grid"
+#if ENABLE_EXPERIMENTAL_FEATURES
+            && tiling_method != "tili" && tiling_method != "unci"
+#endif
+          ) {
+          std::cerr << "Invalid tiling method '" << tiling_method << "'\n";
+          exit(5);
+        }
+        break;
+      case OPTION_UNCI_COMPRESSION: {
+        std::string option(optarg);
+        if (option == "none") {
+          unci_compression = heif_unci_compression_off;
+        }
+        else if (option == "brotli") {
+          unci_compression = heif_unci_compression_brotli;
+        }
+        else if (option == "deflate") {
+          unci_compression = heif_unci_compression_deflate;
+        }
+        else if (option == "zlib") {
+          unci_compression = heif_unci_compression_zlib;
+        }
+        else {
+          std::cerr << "Invalid unci compression method '" << option << "'\n";
+          exit(5);
+        }
+        break;
+      }
       case 'C':
         chroma_downsampling = optarg;
         if (chroma_downsampling != "nn" &&
@@ -638,6 +1021,9 @@ int main(int argc, char** argv)
         }
 #endif
         break;
+      case 'T':
+        use_tiling = true;
+        break;
     }
   }
 
@@ -760,6 +1146,11 @@ int main(int argc, char** argv)
   }
 
 
+  if (lossless && !heif_encoder_descriptor_supports_lossless_compression(active_encoder_descriptor)) {
+    std::cerr << "Warning: the selected encoder does not support lossless encoding. Encoding in lossy mode.\n";
+    lossless = false;
+  }
+
   // If we were given a list of filenames and no '-o' option, check whether the last filename is the desired output filename.
 
   if (output_filename.empty() && argc>1) {
@@ -773,6 +1164,10 @@ int main(int argc, char** argv)
 
   std::shared_ptr<heif_image> primary_image;
 
+  bool is_primary_image = true;
+
+  std::vector<heif_item_id> encoded_image_ids;
+
   for (; optind < argc; optind++) {
     std::string input_filename = argv[optind];
 
@@ -793,44 +1188,37 @@ int main(int argc, char** argv)
 
     // ==============================================================================
 
-    // get file type from file name
-
-    std::string suffix;
-    auto suffix_pos = input_filename.find_last_of('.');
-    if (suffix_pos != std::string::npos) {
-      suffix = input_filename.substr(suffix_pos + 1);
-      std::transform(suffix.begin(), suffix.end(), suffix.begin(), ::tolower);
-    }
-
-    enum
-    {
-      PNG, JPEG, Y4M, TIFF
-    } filetype = JPEG;
-    if (suffix == "png") {
-      filetype = PNG;
-    }
-    else if (suffix == "y4m") {
-      filetype = Y4M;
-    } else if (suffix == "tif" || suffix == "tiff") {
-      filetype = TIFF;
-    }
-
-    InputImage input_image;
-    if (filetype == PNG) {
-      input_image = loadPNG(input_filename.c_str(), output_bit_depth);
-    }
-    else if (filetype == Y4M) {
-      input_image = loadY4M(input_filename.c_str());
-    }
-    else if (filetype == TIFF) {
-      input_image = loadTIFF(input_filename.c_str());
-    }
-    else {
-      input_image = loadJPEG(input_filename.c_str());
-    }
+    InputImage input_image = load_image(input_filename, output_bit_depth);
 
     std::shared_ptr<heif_image> image = input_image.image;
 
+    heif_image_tiling tiling{};
+    std::optional<input_tiles_generator> tile_generator;
+    if (use_tiling) {
+      tile_generator = determine_input_images_tiling(input_filename);
+      if (tile_generator) {
+        tiling.version = 1;
+
+        if (tiled_input_x_y) tile_generator->first_is_x = true;
+
+        tiling.num_columns = tile_generator->nColumns();
+        tiling.num_rows = tile_generator->nRows();
+        tiling.tile_width = heif_image_get_primary_width(image.get());
+        tiling.tile_height = heif_image_get_primary_height(image.get());
+        tiling.image_width = tiling.num_columns * tiling.tile_width;
+        tiling.image_height = tiling.num_rows * tiling.tile_height;
+        tiling.number_of_extra_dimensions = 0;
+      }
+
+      if (tiled_image_width) tiling.image_width = tiled_image_width;
+      if (tiled_image_height) tiling.image_height = tiled_image_height;
+
+      if (!tile_generator || tile_generator->nTiles()==1) {
+        std::cerr << "Cannot enumerate input tiles. Please use filenames with the two tile coordinates in the name.\n";
+        return 5;
+      }
+    }
+
     if (!primary_image) {
       primary_image = image;
     }
@@ -841,83 +1229,14 @@ int main(int argc, char** argv)
     }
 #endif
 
-    heif_color_profile_nclx* nclx = heif_nclx_color_profile_alloc();
-    if (!nclx) {
-      std::cerr << "Cannot allocate NCLX color profile.\n";
-      exit(5);
-    }
-
-    if (lossless) {
-      if (heif_encoder_descriptor_supports_lossless_compression(active_encoder_descriptor)) {
-        heif_encoder_set_lossless(encoder, true);
-
-        if (heif_image_get_colorspace(primary_image.get()) == heif_colorspace_RGB) {
-          nclx->matrix_coefficients = heif_matrix_coefficients_RGB_GBR;
-          nclx->full_range_flag = true;
-          raw_params.emplace_back("chroma=444");
-        }
-        else {
-          heif_color_profile_nclx* input_nclx;
-
-          error = heif_image_get_nclx_color_profile(primary_image.get(), &input_nclx);
-          if (error.code == heif_error_Color_profile_does_not_exist) {
-            // NOP, use default NCLX profile
-          }
-          else if (error.code) {
-            std::cerr << "Cannot get input NCLX color profile.\n";
-            exit(5);
-          }
-          else {
-            nclx->matrix_coefficients = input_nclx->matrix_coefficients;
-            nclx->transfer_characteristics = input_nclx->transfer_characteristics;
-            nclx->color_primaries = input_nclx->color_primaries;
-            nclx->full_range_flag = input_nclx->full_range_flag;
-
-            heif_nclx_color_profile_free(input_nclx);
-          }
-
-          // TODO: this assumes that the encoder plugin has a 'chroma' parameter. Currently, they do, but there should be a better way to set this.
-          switch (heif_image_get_chroma_format(primary_image.get())) {
-            case heif_chroma_420:
-            case heif_chroma_monochrome:
-              raw_params.emplace_back("chroma=420");
-              break;
-            case heif_chroma_422:
-              raw_params.emplace_back("chroma=422");
-              break;
-            case heif_chroma_444:
-              raw_params.emplace_back("chroma=444");
-              break;
-            default:
-              assert(false);
-              exit(5);
-          }
-        }
-      }
-      else {
-        std::cerr << "Warning: the selected encoder does not support lossless encoding. Encoding in lossy mode.\n";
-        lossless = false;
-      }
+    heif_color_profile_nclx* nclx;
+    heif_error error = create_output_nclx_profile_and_configure_encoder(encoder, &nclx, primary_image, lossless);
+    if (error.code) {
+      std::cerr << error.message << "\n";
+      return 5;
     }
 
     if (!lossless) {
-      error = heif_nclx_color_profile_set_matrix_coefficients(nclx, nclx_matrix_coefficients);
-      if (error.code) {
-        std::cerr << "Invalid matrix coefficients specified.\n";
-        exit(5);
-      }
-      error = heif_nclx_color_profile_set_transfer_characteristics(nclx, nclx_transfer_characteristic);
-      if (error.code) {
-        std::cerr << "Invalid transfer characteristics specified.\n";
-        exit(5);
-      }
-      error = heif_nclx_color_profile_set_color_primaries(nclx, nclx_colour_primaries);
-      if (error.code) {
-        std::cerr << "Invalid color primaries specified.\n";
-        exit(5);
-      }
-      nclx->full_range_flag = (uint8_t) nclx_full_range;
-
       heif_encoder_set_lossy_quality(encoder, quality);
     }
 
@@ -943,48 +1262,41 @@ int main(int argc, char** argv)
       options->color_conversion_options.only_use_preferred_chroma_algorithm = true;
     }
 
-    if (crop_to_even_size) {
-      if (heif_image_get_primary_width(image.get()) == 1 ||
-          heif_image_get_primary_height(image.get()) == 1) {
-        std::cerr << "Image only has a size of 1 pixel width or height. Cannot crop to even size.\n";
-        heif_encoder_release(encoder);
-        return 1;
-      }
-
-      std::cerr << "Warning: option --even-size/-E is deprecated as it is not needed anymore.\n";
-
-      int right = heif_image_get_primary_width(image.get()) % 2;
-      int bottom = heif_image_get_primary_height(image.get()) % 2;
-
-      error = heif_image_crop(image.get(), 0, right, 0, bottom);
-      if (error.code != 0) {
-        heif_encoding_options_free(options);
-        heif_nclx_color_profile_free(nclx);
-        heif_encoder_release(encoder);
-        std::cerr << "Could not crop image: " << error.message << "\n";
-        return 1;
-      }
-    }
-
     if (premultiplied_alpha) {
       heif_image_set_premultiplied_alpha(image.get(), premultiplied_alpha);
     }
 
-
     struct heif_image_handle* handle;
-    error = heif_context_encode_image(context.get(),
-                                      image.get(),
-                                      encoder,
-                                      options,
-                                      &handle);
-    if (error.code != 0) {
-      heif_encoding_options_free(options);
-      heif_nclx_color_profile_free(nclx);
-      heif_encoder_release(encoder);
-      std::cerr << "Could not encode HEIF/AVIF file: " << error.message << "\n";
+
+    if (use_tiling) {
+      handle = encode_tiled(context.get(), encoder, options, output_bit_depth, *tile_generator, tiling);
+    }
+    else {
+      error = heif_context_encode_image(context.get(),
+                                        image.get(),
+                                        encoder,
+                                        options,
+                                        &handle);
+      if (error.code != 0) {
+        heif_encoding_options_free(options);
+        heif_nclx_color_profile_free(nclx);
+        heif_encoder_release(encoder);
+        std::cerr << "Could not encode HEIF/AVIF file: " << error.message << "\n";
+        return 1;
+      }
+    }
+
+    if (handle==nullptr) {
+      std::cerr << "Could not encode image\n";
       return 1;
     }
 
+    if (is_primary_image) {
+      heif_context_set_primary_image(context.get(), handle);
+    }
+
+    encoded_image_ids.push_back(heif_image_handle_get_item_id(handle));
+
     // write EXIF to HEIC
     if (!input_image.exif.empty()) {
       // Note: we do not modify the EXIF Orientation here because we want it to match the HEIF transforms.
@@ -1051,6 +1363,8 @@ int main(int argc, char** argv)
     heif_image_handle_release(handle);
     heif_encoding_options_free(options);
     heif_nclx_color_profile_free(nclx);
+
+    is_primary_image = false;
   }
 
   heif_encoder_release(encoder);
@@ -1079,6 +1393,16 @@ int main(int argc, char** argv)
     heif_image_handle_release(primary_image_handle);
   }
 
+#if ENABLE_EXPERIMENTAL_FEATURES
+  if (add_pyramid_group && encoded_image_ids.size() > 1) {
+    error = heif_context_add_pyramid_entity_group(context.get(), encoded_image_ids.data(), encoded_image_ids.size(), nullptr);
+    if (error.code) {
+      std::cerr << "Cannot set multi-resolution pyramid: " << error.message << "\n";
+      return 5;
+    }
+  }
+#endif
+
   error = heif_context_write_to_file(context.get(), output_filename.c_str());
   if (error.code) {
     std::cerr << error.message << "\n";
