@@ -65,10 +65,13 @@ info -w 20012 -o out.265 *file
 info -d // dump
  */
 
+int option_disable_limits = 0;
+
 static struct option long_options[] = {
     //{"write-raw", required_argument, 0, 'w' },
     //{"output",    required_argument, 0, 'o' },
     {(char* const) "dump-boxes", no_argument, 0, 'd'},
+    {(char* const) "disable-limits", no_argument, &option_disable_limits, 1},
     {(char* const) "help",       no_argument, 0, 'h'},
     {(char* const) "version",    no_argument, 0, 'v'},
     {0, 0,                                    0, 0}
@@ -95,6 +98,7 @@ void show_help(const char* argv0)
   //fprintf(stderr,"  -w, --write-raw ID   write raw compressed data of image 'ID'\n");
   //fprintf(stderr,"  -o, --output NAME    output file name for image selected by -w\n");
   fprintf(stderr, "  -d, --dump-boxes     show a low-level dump of all MP4 file boxes\n");
+  fprintf(stderr, "      --disable-limits disable all security limits (do not use in production environment)\n");
   fprintf(stderr, "  -h, --help           show help\n");
   fprintf(stderr, "  -v, --version        show version\n");
 }
@@ -221,19 +225,23 @@ int main(int argc, char** argv)
     return 1;
   }
 
+  if (option_disable_limits) {
+    heif_context_set_security_limits(ctx.get(), heif_get_disabled_security_limits());
+  }
+
   struct heif_error err;
   err = heif_context_read_from_file(ctx.get(), input_filename, nullptr);
 
-  if (dump_boxes) {
-    heif_context_debug_dump_boxes_to_file(ctx.get(), STDOUT_FILENO); // dump to stdout
-    return 0;
-  }
-
   if (err.code != 0) {
     std::cerr << "Could not read HEIF/AVIF file: " << err.message << "\n";
     return 1;
   }
 
+  if (dump_boxes) {
+    heif_context_debug_dump_boxes_to_file(ctx.get(), STDOUT_FILENO); // dump to stdout
+    return 0;
+  }
+
 
   // ==============================================================================
 
@@ -259,6 +267,15 @@ int main(int argc, char** argv)
 
     printf("image: %dx%d (id=%d)%s\n", width, height, IDs[i], primary ? ", primary" : "");
 
+    heif_image_tiling tiling;
+    err = heif_image_handle_get_image_tiling(handle, true, &tiling);
+    if (err.code) {
+      std::cerr << "Error while trying to get image tiling information: " << err.message << "\n";
+    }
+    else if (tiling.num_columns > 0) {
+      std::cout << "  tiles: " << tiling.num_columns << "x" << tiling.num_rows
+                << ", tile size: " << tiling.tile_width << "x" << tiling.tile_height << "\n";
+    }
 
     heif_colorspace colorspace;
     heif_chroma chroma;
@@ -279,6 +296,9 @@ int main(int argc, char** argv)
       case heif_colorspace_monochrome:
         printf("monochrome");
         break;
+      case heif_colorspace_nonvisual:
+        printf("non-visual");
+        break;
       default:
         printf("unknown");
         break;
@@ -672,53 +692,35 @@ int main(int argc, char** argv)
     }
 
 
-    struct heif_image* image;
-    err = heif_decode_image(handle,
-                            &image,
-                            heif_colorspace_undefined,
-                            heif_chroma_undefined,
-                            nullptr);
-    if (err.code) {
-      heif_image_handle_release(handle);
-      std::cerr << "Could not decode image " << err.message << "\n";
-      return 1;
+    uint32_t aspect_h, aspect_v;
+    int has_pasp = heif_image_handle_get_pixel_aspect_ratio(handle, &aspect_h, &aspect_v);
+    if (has_pasp) {
+      std::cout << "pixel aspect ratio: " << aspect_h << "/" << aspect_v << "\n";
     }
 
-    if (image) {
-      uint32_t aspect_h, aspect_v;
-      heif_image_get_pixel_aspect_ratio(image, &aspect_h, &aspect_v);
-      if (aspect_h != aspect_v) {
-        std::cout << "pixel aspect ratio: " << aspect_h << "/" << aspect_v << "\n";
-      }
-
-      if (heif_image_has_content_light_level(image)) {
-        struct heif_content_light_level clli{};
-        heif_image_get_content_light_level(image, &clli);
-        std::cout << "content light level (clli):\n"
-                  << "  max content light level: " << clli.max_content_light_level << "\n"
-                  << "  max pic average light level: " << clli.max_pic_average_light_level << "\n";
-      }
-
-      if (heif_image_has_mastering_display_colour_volume(image)) {
-        struct heif_mastering_display_colour_volume mdcv;
-        heif_image_get_mastering_display_colour_volume(image, &mdcv);
-
-        struct heif_decoded_mastering_display_colour_volume decoded_mdcv;
-        err = heif_mastering_display_colour_volume_decode(&mdcv, &decoded_mdcv);
-
-        std::cout << "mastering display color volume:\n"
-                  << "  display_primaries (x,y): "
-                  << "(" << decoded_mdcv.display_primaries_x[0] << ";" << decoded_mdcv.display_primaries_y[0] << "), "
-                  << "(" << decoded_mdcv.display_primaries_x[1] << ";" << decoded_mdcv.display_primaries_y[1] << "), "
-                  << "(" << decoded_mdcv.display_primaries_x[2] << ";" << decoded_mdcv.display_primaries_y[2] << ")\n";
-
-        std::cout << "  white point (x,y): (" << decoded_mdcv.white_point_x << ";" << decoded_mdcv.white_point_y << ")\n";
-        std::cout << "  max display mastering luminance: " << decoded_mdcv.max_display_mastering_luminance << "\n";
-        std::cout << "  min display mastering luminance: " << decoded_mdcv.min_display_mastering_luminance << "\n";
-      }
+    struct heif_content_light_level clli{};
+    if (heif_image_handle_get_content_light_level(handle, &clli)) {
+      std::cout << "content light level (clli):\n"
+                << "  max content light level: " << clli.max_content_light_level << "\n"
+                << "  max pic average light level: " << clli.max_pic_average_light_level << "\n";
     }
 
-    heif_image_release(image);
+    struct heif_mastering_display_colour_volume mdcv;
+    if (heif_image_handle_get_mastering_display_colour_volume(handle, &mdcv)) {
+
+      struct heif_decoded_mastering_display_colour_volume decoded_mdcv;
+      err = heif_mastering_display_colour_volume_decode(&mdcv, &decoded_mdcv);
+
+      std::cout << "mastering display color volume:\n"
+                << "  display_primaries (x,y): "
+                << "(" << decoded_mdcv.display_primaries_x[0] << ";" << decoded_mdcv.display_primaries_y[0] << "), "
+                << "(" << decoded_mdcv.display_primaries_x[1] << ";" << decoded_mdcv.display_primaries_y[1] << "), "
+                << "(" << decoded_mdcv.display_primaries_x[2] << ";" << decoded_mdcv.display_primaries_y[2] << ")\n";
+
+      std::cout << "  white point (x,y): (" << decoded_mdcv.white_point_x << ";" << decoded_mdcv.white_point_y << ")\n";
+      std::cout << "  max display mastering luminance: " << decoded_mdcv.max_display_mastering_luminance << "\n";
+      std::cout << "  min display mastering luminance: " << decoded_mdcv.min_display_mastering_luminance << "\n";
+    }
 
     heif_image_handle_release(handle);
   }
