@@ -24,10 +24,12 @@
 #include <cstdint>
 #include "common_utils.h"
 #include "libheif/heif.h"
+#include "libheif/heif_experimental.h"
 #include "libheif/heif_properties.h"
 #include <cinttypes>
 #include <cstddef>
 
+#include <utility>
 #include <vector>
 #include <string>
 #include <memory>
@@ -35,6 +37,7 @@
 #include <istream>
 #include <bitset>
 #include <utility>
+#include <optional>
 
 #include "error.h"
 #include "logging.h"
@@ -45,11 +48,6 @@
 #define HAS_BOOL_ARRAY 1
 #endif
 
-// abbreviation
-constexpr inline uint32_t fourcc(const char* id) { return fourcc_to_uint32(id); }
-
-std::string to_fourcc(uint32_t code);
-
 /*
   constexpr uint32_t fourcc(const char* string)
   {
@@ -157,6 +155,22 @@ protected:
 };
 
 
+enum class BoxStorageMode {
+  Undefined,
+  Memory,
+  Parsed,
+  File
+};
+
+
+// Consequence of a box parse error
+enum class parse_error_fatality {
+  fatal,     // failure to parse this box leads to the associated item being unreable
+  ignorable, // ignoring this box will lead to unexpected output, but may be better than nothing
+  optional   // the box contains extra information that is not essential for viewing
+};
+
+
 class Box : public BoxHeader
 {
 public:
@@ -170,7 +184,7 @@ public:
   // header size without the FullBox fields (if applicable)
   int calculate_header_size(bool data64bit) const;
 
-  static Error read(BitstreamRange& range, std::shared_ptr<Box>* box);
+  static Error read(BitstreamRange& range, std::shared_ptr<Box>* box, const heif_security_limits*);
 
   virtual Error write(StreamWriter& writer) const;
 
@@ -181,42 +195,78 @@ public:
 
   std::string dump(Indent&) const override;
 
-  std::shared_ptr<Box> get_child_box(uint32_t short_type) const;
+  template<typename T> [[nodiscard]] std::shared_ptr<T> get_child_box() const
+  {
+    // TODO: we could remove the dynamic_cast<> by adding the fourcc type of each Box
+    //       as a "constexpr uint32_t Box::short_type", compare to that and use static_cast<>
+    for (auto& box : m_children) {
+      if (auto typed_box = std::dynamic_pointer_cast<T>(box)) {
+        return typed_box;
+      }
+    }
+
+    return nullptr;
+  }
+
 
-  std::vector<std::shared_ptr<Box>> get_child_boxes(uint32_t short_type) const;
 
   template<typename T>
-  std::vector<std::shared_ptr<T>> get_typed_child_boxes(uint32_t short_type) const
+  std::vector<std::shared_ptr<T>> get_child_boxes() const
   {
-    auto boxes = get_child_boxes(short_type);
-    std::vector<std::shared_ptr<T>> typedBoxes;
-    for (const auto& box : boxes) {
-      typedBoxes.push_back(std::dynamic_pointer_cast<T>(box));
+    std::vector<std::shared_ptr<T>> result;
+    for (auto& box : m_children) {
+      if (auto typed_box = std::dynamic_pointer_cast<T>(box)) {
+        result.push_back(typed_box);
+      }
     }
-    return typedBoxes;
+
+    return result;
   }
 
   const std::vector<std::shared_ptr<Box>>& get_all_child_boxes() const { return m_children; }
 
-  int append_child_box(const std::shared_ptr<Box>& box)
+  uint32_t append_child_box(const std::shared_ptr<Box>& box)
   {
     m_children.push_back(box);
     return (int) m_children.size() - 1;
   }
 
+  bool has_child_boxes() const { return !m_children.empty(); }
+
   virtual bool operator==(const Box& other) const;
 
   static bool equal(const std::shared_ptr<Box>& box1, const std::shared_ptr<Box>& box2);
 
+  BoxStorageMode get_box_storage_mode() const { return m_storage_mode; }
+
+  void set_output_position(uint64_t pos) { m_output_position = pos; }
+
+  virtual parse_error_fatality get_parse_error_fatality() const { return parse_error_fatality::fatal; }
+
+  virtual bool is_essential() const { return m_is_essential; } // only used for properties
+
+  void set_is_essential(bool flag) { m_is_essential = flag; }
 
 protected:
-  virtual Error parse(BitstreamRange& range);
+  virtual Error parse(BitstreamRange& range, const heif_security_limits* limits);
 
   std::vector<std::shared_ptr<Box>> m_children;
 
-  const static int READ_CHILDREN_ALL = -1;
+  BoxStorageMode m_storage_mode = BoxStorageMode::Undefined;
 
-  Error read_children(BitstreamRange& range, int number = READ_CHILDREN_ALL);
+  const static uint64_t INVALID_POSITION = 0xFFFFFFFFFFFFFFFF;
+
+  uint64_t m_input_position = INVALID_POSITION;
+  uint64_t m_output_position = INVALID_POSITION;
+  std::vector<uint8_t> m_box_data; // including header
+
+  std::string m_debug_box_type;
+
+  bool m_is_essential = false;
+
+  const static uint32_t READ_CHILDREN_ALL = 0xFFFFFFFF;
+
+  Error read_children(BitstreamRange& range, uint32_t number /* READ_CHILDREN_ALL */, const heif_security_limits* limits);
 
   Error write_children(StreamWriter& writer) const;
 
@@ -286,13 +336,43 @@ public:
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   std::vector<uint8_t> m_data;
 };
 
 
 
+class Box_Error : public Box
+{
+public:
+  Box_Error(uint32_t box4cc, Error err, parse_error_fatality fatality)
+  {
+    set_short_type(fourcc("ERR "));
+
+    m_box_type_with_parse_error = box4cc;
+    m_error = std::move(err);
+    m_fatality = fatality;
+  }
+
+  Error write(StreamWriter& writer) const override { return {heif_error_Usage_error, heif_suberror_Unspecified, "Cannot write dummy error box."}; }
+
+  std::string dump(Indent&) const override;
+
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override;
+
+  [[nodiscard]] Error get_error() const { return m_error; }
+
+protected:
+  Error parse(BitstreamRange& range, const heif_security_limits*) override { assert(false); return Error::Ok; }
+
+  uint32_t m_box_type_with_parse_error;
+  Error m_error;
+  parse_error_fatality m_fatality;
+};
+
+
+
 
 class Box_ftyp : public Box
 {
@@ -308,8 +388,12 @@ public:
 
   std::vector<uint32_t> list_brands() const { return m_compatible_brands; }
 
+  uint32_t get_major_brand() const { return m_major_brand; }
+
   void set_major_brand(heif_brand2 major_brand) { m_major_brand = major_brand; }
 
+  uint32_t get_minor_version() const { return m_minor_version; }
+
   void set_minor_version(uint32_t minor_version) { m_minor_version = minor_version; }
 
   void clear_compatible_brands() { m_compatible_brands.clear(); }
@@ -319,7 +403,7 @@ public:
   Error write(StreamWriter& writer) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   uint32_t m_major_brand = 0;
@@ -339,7 +423,7 @@ public:
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -362,7 +446,7 @@ public:
   void set_name(std::string name) { m_name = std::move(name); }
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   uint32_t m_pre_defined = 0;
@@ -391,7 +475,7 @@ public:
   Error write(StreamWriter& writer) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   heif_item_id m_item_ID = 0;
@@ -401,10 +485,11 @@ private:
 class Box_iloc : public FullBox
 {
 public:
-  Box_iloc()
-  {
-    set_short_type(fourcc("iloc"));
-  }
+  Box_iloc();
+
+  ~Box_iloc() override;
+
+  void set_use_tmp_file(bool flag);
 
   std::string dump(Indent&) const override;
 
@@ -429,18 +514,33 @@ public:
 
   const std::vector<Item>& get_items() const { return m_items; }
 
-  Error read_data(const Item& item,
+  Error read_data(heif_item_id item,
                   const std::shared_ptr<StreamReader>& istr,
                   const std::shared_ptr<class Box_idat>&,
-                  std::vector<uint8_t>* dest) const;
+                  std::vector<uint8_t>* dest,
+                  const heif_security_limits* limits) const;
+
+  // Note: size==std::numeric_limits<uint64_t>::max() reads the data until the end
+  Error read_data(heif_item_id item,
+                  const std::shared_ptr<StreamReader>& istr,
+                  const std::shared_ptr<class Box_idat>&,
+                  std::vector<uint8_t>* dest,
+                  uint64_t offset, uint64_t size,
+                  const heif_security_limits* limits) const;
 
   void set_min_version(uint8_t min_version) { m_user_defined_min_version = min_version; }
 
   // append bitstream data that will be written later (after iloc box)
+  // TODO: use an enum for the construction method
   Error append_data(heif_item_id item_ID,
                     const std::vector<uint8_t>& data,
                     uint8_t construction_method = 0);
 
+  Error replace_data(heif_item_id item_ID,
+                     uint64_t output_offset,
+                     const std::vector<uint8_t>& data,
+                     uint8_t construction_method);
+
   // append bitstream data that already has been written (before iloc box)
   // Error write_mdat_before_iloc(heif_image_id item_ID,
   //                              std::vector<uint8_t>& data)
@@ -457,8 +557,10 @@ public:
 
   Error write_mdat_after_iloc(StreamWriter& writer);
 
+  void append_item(Item &item) { m_items.push_back(item); }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   std::vector<Item> m_items;
@@ -473,6 +575,10 @@ private:
   void patch_iloc_header(StreamWriter& writer) const;
 
   int m_idat_offset = 0; // only for writing: offset of next data array
+
+  bool m_use_tmpfile = false;
+  int m_tmpfile_fd = 0;
+  char m_tmp_filename[20];
 };
 
 
@@ -494,9 +600,9 @@ public:
 
   void set_item_ID(heif_item_id id) { m_item_ID = id; }
 
-  const std::string& get_item_type() const { return m_item_type; }
+  uint32_t get_item_type_4cc() const { return m_item_type_4cc; }
 
-  void set_item_type(const std::string& type) { m_item_type = type; }
+  void set_item_type_4cc(uint32_t type) { m_item_type_4cc = type; }
 
   void set_item_name(const std::string& name) { m_item_name = name; }
 
@@ -519,13 +625,13 @@ public:
   void set_item_uri_type(const std::string& uritype) { m_item_uri_type = uritype; }
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   heif_item_id m_item_ID = 0;
   uint16_t m_item_protection_index = 0;
 
-  std::string m_item_type;
+  uint32_t m_item_type_4cc = 0;
   std::string m_item_name;
   std::string m_content_type;
   std::string m_content_encoding;
@@ -551,7 +657,7 @@ public:
   Error write(StreamWriter& writer) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   //std::vector< std::shared_ptr<Box_infe> > m_iteminfos;
@@ -569,7 +675,7 @@ public:
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -581,7 +687,7 @@ public:
     set_short_type(fourcc("ipco"));
   }
 
-  int find_or_append_child_box(const std::shared_ptr<Box>& box);
+  uint32_t find_or_append_child_box(const std::shared_ptr<Box>& box);
 
   Error get_properties_for_item_ID(heif_item_id itemID,
                                    const std::shared_ptr<class Box_ipma>&,
@@ -598,7 +704,7 @@ public:
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -626,8 +732,10 @@ public:
 
   bool operator==(const Box& other) const override;
 
+  bool is_essential() const override { return false; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   uint32_t m_image_width = 0;
@@ -665,7 +773,7 @@ public:
   void insert_entries_from_other_ipma_box(const Box_ipma& b);
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   struct Entry
   {
@@ -691,10 +799,12 @@ public:
 
   const std::vector<uint8_t>& get_subtypes() const { return m_aux_subtypes; }
 
+  bool is_essential() const override { return true; }
+
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
@@ -712,15 +822,19 @@ public:
     set_short_type(fourcc("irot"));
   }
 
+  bool is_essential() const override { return true; }
+
   std::string dump(Indent&) const override;
 
-  int get_rotation() const { return m_rotation; }
+  int get_rotation_ccw() const { return m_rotation; }
 
   // Only these multiples of 90 are allowed: 0, 90, 180, 270.
   void set_rotation_ccw(int rot) { m_rotation = rot; }
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::ignorable; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
@@ -737,14 +851,18 @@ public:
     set_short_type(fourcc("imir"));
   }
 
+  bool is_essential() const override { return true; }
+
   heif_transform_mirror_direction get_mirror_direction() const { return m_axis; }
 
   void set_mirror_direction(heif_transform_mirror_direction dir) { m_axis = dir; }
 
   std::string dump(Indent&) const override;
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::ignorable; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
@@ -761,12 +879,14 @@ public:
     set_short_type(fourcc("clap"));
   }
 
+  bool is_essential() const override { return true; }
+
   std::string dump(Indent&) const override;
 
-  int left_rounded(int image_width) const;  // first column
-  int right_rounded(int image_width) const; // last column that is part of the cropped image
-  int top_rounded(int image_height) const;   // first row
-  int bottom_rounded(int image_height) const; // last row included in the cropped image
+  int left_rounded(uint32_t image_width) const;  // first column
+  int right_rounded(uint32_t image_width) const; // last column that is part of the cropped image
+  int top_rounded(uint32_t image_height) const;   // first row
+  int bottom_rounded(uint32_t image_height) const; // last row included in the cropped image
 
   double left(int image_width) const;
   double top(int image_height) const;
@@ -778,8 +898,10 @@ public:
   void set(uint32_t clap_width, uint32_t clap_height,
            uint32_t image_width, uint32_t image_height);
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::ignorable; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
@@ -818,13 +940,17 @@ public:
 
   void add_references(heif_item_id from_id, uint32_t type, const std::vector<heif_item_id>& to_ids);
 
+  void overwrite_reference(heif_item_id from_id, uint32_t type, uint32_t reference_idx, heif_item_id to_item);
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
   void derive_box_version() override;
 
+  Error check_for_double_references() const;
+
 private:
   std::vector<Reference> m_references;
 };
@@ -837,7 +963,8 @@ public:
 
   Error read_data(const std::shared_ptr<StreamReader>& istr,
                   uint64_t start, uint64_t length,
-                  std::vector<uint8_t>& out_data) const;
+                  std::vector<uint8_t>& out_data,
+                  const heif_security_limits* limits) const;
 
   int append_data(const std::vector<uint8_t>& data)
   {
@@ -853,7 +980,7 @@ public:
   Error write(StreamWriter& writer) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   std::streampos m_data_start_pos;
 
@@ -864,10 +991,15 @@ protected:
 class Box_grpl : public Box
 {
 public:
+  Box_grpl()
+  {
+    set_short_type(fourcc("grpl"));
+  }
+
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -876,17 +1008,34 @@ class Box_EntityToGroup : public FullBox
 public:
   std::string dump(Indent&) const override;
 
+  Error write(StreamWriter& writer) const override;
+
+  void set_group_id(heif_entity_group_id id) { group_id = id; }
+
+  heif_entity_group_id get_group_id() const { return group_id; }
+
+  void set_item_ids(const std::vector<heif_item_id>& ids) { entity_ids = ids; }
+
+  const std::vector<heif_item_id>& get_item_ids() const { return entity_ids; }
+
 protected:
-  uint32_t group_id;
+  heif_entity_group_id group_id = 0;
   std::vector<heif_item_id> entity_ids;
 
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
+
+  void write_entity_group_ids(StreamWriter& writer) const;
 };
 
 
 class Box_ster : public Box_EntityToGroup
 {
 public:
+  Box_ster()
+  {
+    set_short_type(fourcc("ster"));
+  }
+
   std::string dump(Indent&) const override;
 
   heif_item_id get_left_image() const { return entity_ids[0]; }
@@ -894,18 +1043,21 @@ public:
 
 protected:
 
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
 class Box_pymd : public Box_EntityToGroup
 {
 public:
+  Box_pymd()
+  {
+    set_short_type(fourcc("pymd"));
+  }
+
   std::string dump(Indent&) const override;
 
-protected:
-  uint16_t tile_size_x;
-  uint16_t tile_size_y;
+  Error write(StreamWriter& writer) const override;
 
   struct LayerInfo {
     uint16_t layer_binning;
@@ -913,9 +1065,28 @@ protected:
     uint16_t tiles_in_layer_column_minus1;
   };
 
+  void set_layers(uint16_t _tile_size_x,
+                  uint16_t _tile_size_y,
+                  const std::vector<LayerInfo>& layers,
+                  const std::vector<heif_item_id>& layer_item_ids) // low to high resolution
+  {
+    set_item_ids(layer_item_ids);
+    m_layer_infos = layers;
+    tile_size_x = _tile_size_x;
+    tile_size_y = _tile_size_y;
+  }
+
+  const std::vector<LayerInfo>& get_layers() const { return m_layer_infos; }
+
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::ignorable; }
+
+protected:
+  uint16_t tile_size_x = 0;
+  uint16_t tile_size_y = 0;
+
   std::vector<LayerInfo> m_layer_infos;
 
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -927,7 +1098,7 @@ public:
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -937,7 +1108,7 @@ public:
   std::string dump(Indent&) const override;
 
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -946,8 +1117,10 @@ class Box_url : public FullBox
 public:
   std::string dump(Indent&) const override;
 
+  bool is_same_file() const { return m_location.empty(); }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   std::string m_location;
 };
@@ -973,8 +1146,10 @@ public:
 
   Error write(StreamWriter& writer) const override;
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   std::vector<uint8_t> m_bits_per_channel;
@@ -996,8 +1171,10 @@ public:
 
   Error write(StreamWriter& writer) const override;
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -1015,8 +1192,10 @@ public:
 
   Error write(StreamWriter& writer) const override;
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -1037,8 +1216,10 @@ public:
 
   Error write(StreamWriter& writer) const override;
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 };
 
 
@@ -1053,8 +1234,74 @@ public:
 
   Error write(StreamWriter& writer) const override;
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
+};
+
+
+class Box_amve : public Box
+{
+public:
+  Box_amve();
+
+  heif_ambient_viewing_environment amve;
+
+  std::string dump(Indent&) const override;
+
+  Error write(StreamWriter& writer) const override;
+
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
+protected:
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
+};
+
+
+class Box_cclv : public Box
+{
+public:
+  Box_cclv();
+
+  bool ccv_primaries_are_valid() const { return m_ccv_primaries_valid; }
+  int32_t get_ccv_primary_x0() const { return m_ccv_primaries_x[0]; }
+  int32_t get_ccv_primary_y0() const { return m_ccv_primaries_y[0]; }
+  int32_t get_ccv_primary_x1() const { return m_ccv_primaries_x[1]; }
+  int32_t get_ccv_primary_y1() const { return m_ccv_primaries_y[1]; }
+  int32_t get_ccv_primary_x2() const { return m_ccv_primaries_x[2]; }
+  int32_t get_ccv_primary_y2() const { return m_ccv_primaries_y[2]; }
+  void set_primaries(int32_t x0, int32_t y0, int32_t x1, int32_t y1, int32_t x2, int32_t y2);
+
+  bool min_luminance_is_valid() const { return m_ccv_min_luminance_value.has_value(); }
+  uint32_t get_min_luminance() const { return *m_ccv_min_luminance_value; }
+  void set_min_luminance(uint32_t luminance) { m_ccv_min_luminance_value = luminance; }
+
+  bool max_luminance_is_valid() const { return m_ccv_max_luminance_value.has_value(); }
+  uint32_t get_max_luminance() const { return *m_ccv_max_luminance_value; }
+  void set_max_luminance(uint32_t luminance) { m_ccv_max_luminance_value = luminance; }
+
+  bool avg_luminance_is_valid() const { return m_ccv_avg_luminance_value.has_value(); }
+  uint32_t get_avg_luminance() const { return *m_ccv_avg_luminance_value; }
+  void set_avg_luminance(uint32_t luminance) { m_ccv_avg_luminance_value = luminance; }
+
+  std::string dump(Indent&) const override;
+
+  Error write(StreamWriter& writer) const override;
+
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
+protected:
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
+
+private:
+  bool m_ccv_primaries_valid = false;
+  int32_t m_ccv_primaries_x[3] {};
+  int32_t m_ccv_primaries_y[3] {};
+
+  std::optional<uint32_t> m_ccv_min_luminance_value;
+  std::optional<uint32_t> m_ccv_max_luminance_value;
+  std::optional<uint32_t> m_ccv_avg_luminance_value;
 };
 
 
@@ -1123,8 +1370,10 @@ public:
 
   void set_intrinsic_matrix(RelativeIntrinsicMatrix matrix);
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
@@ -1176,8 +1425,10 @@ public:
 
   Error set_extrinsic_matrix(ExtrinsicMatrix matrix);
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
   Error write(StreamWriter& writer) const override;
 
@@ -1286,8 +1537,10 @@ public:
    */
   void set_tags(const std::string tags) { m_tags = tags; }
 
+  [[nodiscard]] parse_error_fatality get_parse_error_fatality() const override { return parse_error_fatality::optional; }
+
 protected:
-  Error parse(BitstreamRange& range) override;
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
 
 private:
   std::string m_lang;
@@ -1296,4 +1549,121 @@ private:
   std::string m_tags;
 };
 
+
+#if ENABLE_EXPERIMENTAL_FEATURES
+class Box_taic : public FullBox
+{
+public:
+  Box_taic()
+  {
+    set_short_type(fourcc("taic"));
+  }
+
+  std::string dump(Indent&) const override;
+
+  Error write(StreamWriter& writer) const override;
+
+  /**
+   * time_uncertainty.
+   * 
+   * The standard deviation measurement uncertainty in nanoseconds
+   * for the timestamp generation process. 
+   */
+  void set_time_uncertainty(uint64_t time_uncertainty) { m_time_uncertainty = time_uncertainty;}
+  
+  /**
+   * clock_resolution.
+   * 
+   * Specifies the resolution of the receptor clock in nanoseconds.
+   * For example, a microsecond clock has a clock_resolution of 1000.
+   */
+  void set_clock_resolution(uint32_t clock_resolution) { m_clock_resolution = clock_resolution; }
+  
+  /**
+   * clock_drift_rate.
+   * 
+   * The difference between the synchronized and unsynchronized
+   * time, over a period of one second. 
+   */
+  void set_clock_drift_rate(int32_t clock_drift_rate) { m_clock_drift_rate = clock_drift_rate; }
+  
+  /**
+   * clock_type.
+   * 
+   * 0 = Clock type is unkown
+   * 1 = The clock does not synchronize to an atomic source of absolute TAI time
+   * 2 = The clock can synchronize to an atomic source of absolute TAI time
+   */
+  void set_clock_type(uint8_t clock_type) { m_clock_type = clock_type; }
+
+  uint64_t get_time_uncertainty() const { return m_time_uncertainty; }
+  
+  uint32_t get_clock_resolution() const { return m_clock_resolution; }
+  
+  int32_t get_clock_drift_rate() const { return m_clock_drift_rate; }
+  
+  uint8_t get_clock_type() const { return m_clock_type; }
+
+protected:
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
+
+private:
+  uint64_t m_time_uncertainty = heif_tai_clock_info_unknown_time_uncertainty;
+  uint32_t m_clock_resolution = 0;
+  int32_t m_clock_drift_rate = heif_tai_clock_info_unknown_drift_rate;
+  uint8_t m_clock_type = 0;
+};
+
+
+class Box_itai : public FullBox 
+{
+public:
+  Box_itai()
+  {
+    set_short_type(fourcc("itai"));
+  }
+
+  std::string dump(Indent&) const override;
+
+  Error write(StreamWriter& writer) const override;
+
+  /**
+   * The number of nanoseconds since the TAI epoch of 1958-01-01T00:00:00.0Z.
+   */
+  void set_tai_timestamp(uint64_t timestamp) { m_tai_timestamp = timestamp; }
+
+  /**
+  * synchronization_state (0=unsynchronized, 1=synchronized)
+  */
+  void set_synchronization_state(bool state) { m_synchronization_state = state; }
+
+  /**
+  * timestamp_generation_failure (0=generated, 1=failed)
+  */
+  void set_timestamp_generation_failure(bool failure) { m_timestamp_generation_failure = failure; }
+
+  /**
+   * timestamp_is_modified (0=original 1=modified)
+   */
+  void set_timestamp_is_modified(bool is_modified) { m_timestamp_is_modified = is_modified; }
+
+  uint64_t get_tai_timestamp() const { return m_tai_timestamp; }
+
+  bool get_synchronization_state() const { return m_synchronization_state; }
+
+  bool get_timestamp_generation_failure() const { return m_timestamp_generation_failure; }
+
+  bool get_timestamp_is_modified() const { return m_timestamp_is_modified; }
+
+protected:
+  Error parse(BitstreamRange& range, const heif_security_limits*) override;
+
+private:
+  uint64_t m_tai_timestamp = 0;
+  bool m_synchronization_state = false;
+  bool m_timestamp_generation_failure = false;
+  bool m_timestamp_is_modified = false;
+};
 #endif
+
+#endif
\ No newline at end of file
