@@ -48,7 +48,7 @@ Op_RGB_to_RGB24_32::state_after_conversion(const ColorState& input_state,
   output_state.has_alpha = true;
   output_state.bits_per_pixel = 8;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   // --- convert to RGB (without alpha)
 
@@ -57,7 +57,7 @@ Op_RGB_to_RGB24_32::state_after_conversion(const ColorState& input_state,
   output_state.has_alpha = false;
   output_state.bits_per_pixel = 8;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   return states;
 }
@@ -84,8 +84,8 @@ Op_RGB_to_RGB24_32::convert_colorspace(const std::shared_ptr<const HeifPixelImag
 
   auto outimg = std::make_shared<HeifPixelImage>();
 
-  int width = input->get_width();
-  int height = input->get_height();
+  uint32_t width = input->get_width();
+  uint32_t height = input->get_height();
 
   outimg->create(width, height, heif_colorspace_RGB,
                  want_alpha ? heif_chroma_interleaved_32bit : heif_chroma_interleaved_24bit);
@@ -95,10 +95,10 @@ Op_RGB_to_RGB24_32::convert_colorspace(const std::shared_ptr<const HeifPixelImag
   }
 
   const uint8_t* in_r, * in_g, * in_b, * in_a = nullptr;
-  int in_r_stride = 0, in_g_stride = 0, in_b_stride = 0, in_a_stride = 0;
+  uint32_t in_r_stride = 0, in_g_stride = 0, in_b_stride = 0, in_a_stride = 0;
 
   uint8_t* out_p;
-  int out_p_stride = 0;
+  uint32_t out_p_stride = 0;
 
   in_r = input->get_plane(heif_channel_R, &in_r_stride);
   in_g = input->get_plane(heif_channel_G, &in_g_stride);
@@ -109,7 +109,7 @@ Op_RGB_to_RGB24_32::convert_colorspace(const std::shared_ptr<const HeifPixelImag
     in_a = input->get_plane(heif_channel_Alpha, &in_a_stride);
   }
 
-  int x, y;
+  uint32_t x, y;
   for (y = 0; y < height; y++) {
 
     if (has_alpha && want_alpha) {
@@ -168,7 +168,7 @@ Op_RGB_HDR_to_RRGGBBaa_BE::state_after_conversion(const ColorState& input_state,
     output_state.has_alpha = false;
     output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-    states.push_back({output_state, SpeedCosts_Unoptimized});
+    states.emplace_back(output_state, SpeedCosts_Unoptimized);
   }
 
 
@@ -179,7 +179,7 @@ Op_RGB_HDR_to_RRGGBBaa_BE::state_after_conversion(const ColorState& input_state,
   output_state.has_alpha = true;
   output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
 
   return states;
@@ -216,8 +216,8 @@ Op_RGB_HDR_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPi
 
   auto outimg = std::make_shared<HeifPixelImage>();
 
-  int width = input->get_width();
-  int height = input->get_height();
+  uint32_t width = input->get_width();
+  uint32_t height = input->get_height();
 
   outimg->create(width, height, heif_colorspace_RGB,
                  output_has_alpha ? heif_chroma_interleaved_RRGGBBAA_BE : heif_chroma_interleaved_RRGGBB_BE);
@@ -226,10 +226,10 @@ Op_RGB_HDR_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPi
   }
 
   const uint16_t* in_r, * in_g, * in_b, * in_a = nullptr;
-  int in_r_stride = 0, in_g_stride = 0, in_b_stride = 0, in_a_stride = 0;
+  uint32_t in_r_stride = 0, in_g_stride = 0, in_b_stride = 0, in_a_stride = 0;
 
   uint8_t* out_p;
-  int out_p_stride = 0;
+  uint32_t out_p_stride = 0;
 
   in_r = (uint16_t*) input->get_plane(heif_channel_R, &in_r_stride);
   in_g = (uint16_t*) input->get_plane(heif_channel_G, &in_g_stride);
@@ -238,6 +238,12 @@ Op_RGB_HDR_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPi
 
   if (input_has_alpha) {
     in_a = (uint16_t*) input->get_plane(heif_channel_Alpha, &in_a_stride);
+    assert(in_a != nullptr);
+
+    // should never happen, but makes clang-tidy happy
+    if (in_a == nullptr) {
+      return nullptr;
+    }
   }
 
   in_r_stride /= 2;
@@ -247,8 +253,8 @@ Op_RGB_HDR_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPi
 
   const int pixelsize = (output_has_alpha ? 8 : 6);
 
-  int x, y;
-  uint16_t alpha_max = static_cast<uint16_t>((1 << bpp) - 1);
+  uint32_t x, y;
+  auto alpha_max = static_cast<uint16_t>((1 << bpp) - 1);
   for (y = 0; y < height; y++) {
     for (x = 0; x < width; x++) {
       uint16_t r = in_r[x + y * in_r_stride];
@@ -297,7 +303,7 @@ Op_RGB_to_RRGGBBaa_BE::state_after_conversion(const ColorState& input_state,
     output_state.has_alpha = false;
     output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-    states.push_back({output_state, SpeedCosts_Unoptimized});
+    states.emplace_back(output_state, SpeedCosts_Unoptimized);
   }
 
 
@@ -308,8 +314,7 @@ Op_RGB_to_RRGGBBaa_BE::state_after_conversion(const ColorState& input_state,
   output_state.has_alpha = true;
   output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
-
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   return states;
 }
@@ -338,8 +343,8 @@ Op_RGB_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPixelI
 
   auto outimg = std::make_shared<HeifPixelImage>();
 
-  int width = input->get_width();
-  int height = input->get_height();
+  uint32_t width = input->get_width();
+  uint32_t height = input->get_height();
 
   outimg->create(width, height, heif_colorspace_RGB,
                  output_has_alpha ? heif_chroma_interleaved_RRGGBBAA_BE : heif_chroma_interleaved_RRGGBB_BE);
@@ -349,10 +354,10 @@ Op_RGB_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPixelI
   }
 
   const uint8_t* in_r, * in_g, * in_b, * in_a = nullptr;
-  int in_r_stride = 0, in_g_stride = 0, in_b_stride = 0, in_a_stride = 0;
+  uint32_t in_r_stride = 0, in_g_stride = 0, in_b_stride = 0, in_a_stride = 0;
 
   uint8_t* out_p;
-  int out_p_stride = 0;
+  uint32_t out_p_stride = 0;
 
   in_r = input->get_plane(heif_channel_R, &in_r_stride);
   in_g = input->get_plane(heif_channel_G, &in_g_stride);
@@ -365,7 +370,7 @@ Op_RGB_to_RRGGBBaa_BE::convert_colorspace(const std::shared_ptr<const HeifPixelI
 
   const int pixelsize = (output_has_alpha ? 8 : 6);
 
-  int x, y;
+  uint32_t x, y;
   for (y = 0; y < height; y++) {
 
     if (input_has_alpha) {
@@ -426,8 +431,7 @@ Op_RRGGBBaa_BE_to_RGB_HDR::state_after_conversion(const ColorState& input_state,
   output_state.has_alpha = target_state.has_alpha;
   output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
-
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   return states;
 }
@@ -445,8 +449,8 @@ Op_RRGGBBaa_BE_to_RGB_HDR::convert_colorspace(const std::shared_ptr<const HeifPi
 
   auto outimg = std::make_shared<HeifPixelImage>();
 
-  int width = input->get_width();
-  int height = input->get_height();
+  uint32_t width = input->get_width();
+  uint32_t height = input->get_height();
   int bpp = input->get_bits_per_pixel(heif_channel_interleaved);
 
   outimg->create(width, height, heif_colorspace_RGB, heif_chroma_444);
@@ -464,11 +468,11 @@ Op_RRGGBBaa_BE_to_RGB_HDR::convert_colorspace(const std::shared_ptr<const HeifPi
   }
 
   const uint8_t* in_p;
-  int in_p_stride = 0;
+  uint32_t in_p_stride = 0;
   int in_pix_size = has_alpha ? 8 : 6;
 
   uint16_t* out_r, * out_g, * out_b, * out_a = nullptr;
-  int out_r_stride = 0, out_g_stride = 0, out_b_stride = 0, out_a_stride = 0;
+  uint32_t out_r_stride = 0, out_g_stride = 0, out_b_stride = 0, out_a_stride = 0;
 
   in_p = input->get_plane(heif_channel_interleaved, &in_p_stride);
 
@@ -485,8 +489,8 @@ Op_RRGGBBaa_BE_to_RGB_HDR::convert_colorspace(const std::shared_ptr<const HeifPi
   out_b_stride /= 2;
   out_a_stride /= 2;
 
-  uint16_t alpha_max = static_cast<uint16_t>((1 << bpp) - 1);
-  int x, y;
+  auto alpha_max = static_cast<uint16_t>((1 << bpp) - 1);
+  uint32_t x, y;
   for (y = 0; y < height; y++) {
 
     for (x = 0; x < width; x++) {
@@ -540,7 +544,7 @@ Op_RGB24_32_to_RGB::state_after_conversion(const ColorState& input_state,
   output_state.has_alpha = target_state.has_alpha;
   output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   return states;
 }
@@ -557,8 +561,8 @@ Op_RGB24_32_to_RGB::convert_colorspace(const std::shared_ptr<const HeifPixelImag
 
   auto outimg = std::make_shared<HeifPixelImage>();
 
-  int width = input->get_width();
-  int height = input->get_height();
+  uint32_t width = input->get_width();
+  uint32_t height = input->get_height();
 
   outimg->create(width, height, heif_colorspace_RGB, heif_chroma_444);
 
@@ -575,11 +579,11 @@ Op_RGB24_32_to_RGB::convert_colorspace(const std::shared_ptr<const HeifPixelImag
   }
 
   const uint8_t* in_p;
-  int in_p_stride = 0;
+  uint32_t in_p_stride = 0;
   int in_pix_size = has_alpha ? 4 : 3;
 
   uint8_t* out_r, * out_g, * out_b, * out_a = nullptr;
-  int out_r_stride = 0, out_g_stride = 0, out_b_stride = 0, out_a_stride = 0;
+  uint32_t out_r_stride = 0, out_g_stride = 0, out_b_stride = 0, out_a_stride = 0;
 
   in_p = input->get_plane(heif_channel_interleaved, &in_p_stride);
 
@@ -591,7 +595,7 @@ Op_RGB24_32_to_RGB::convert_colorspace(const std::shared_ptr<const HeifPixelImag
     out_a = outimg->get_plane(heif_channel_Alpha, &out_a_stride);
   }
 
-  int x, y;
+  uint32_t x, y;
   for (y = 0; y < height; y++) {
 
     for (x = 0; x < width; x++) {
@@ -645,7 +649,7 @@ Op_RRGGBBaa_swap_endianness::state_after_conversion(const ColorState& input_stat
     output_state.has_alpha = false;
     output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-    states.push_back({output_state, SpeedCosts_Unoptimized});
+    states.emplace_back(output_state, SpeedCosts_Unoptimized);
   }
 
 
@@ -665,10 +669,9 @@ Op_RRGGBBaa_swap_endianness::state_after_conversion(const ColorState& input_stat
     output_state.has_alpha = true;
     output_state.bits_per_pixel = input_state.bits_per_pixel;
 
-    states.push_back({output_state, SpeedCosts_Unoptimized});
+    states.emplace_back(output_state, SpeedCosts_Unoptimized);
   }
 
-
   return states;
 }
 
@@ -681,8 +684,8 @@ Op_RRGGBBaa_swap_endianness::convert_colorspace(const std::shared_ptr<const Heif
 {
   auto outimg = std::make_shared<HeifPixelImage>();
 
-  int width = input->get_width();
-  int height = input->get_height();
+  uint32_t width = input->get_width();
+  uint32_t height = input->get_height();
 
   switch (input->get_chroma_format()) {
     case heif_chroma_interleaved_RRGGBB_LE:
@@ -707,17 +710,17 @@ Op_RRGGBBaa_swap_endianness::convert_colorspace(const std::shared_ptr<const Heif
   }
 
   const uint8_t* in_p = nullptr;
-  int in_p_stride = 0;
+  uint32_t in_p_stride = 0;
 
   uint8_t* out_p;
-  int out_p_stride = 0;
+  uint32_t out_p_stride = 0;
 
   in_p = input->get_plane(heif_channel_interleaved, &in_p_stride);
   out_p = outimg->get_plane(heif_channel_interleaved, &out_p_stride);
 
-  int n_bytes = std::min(in_p_stride, out_p_stride);
+  uint32_t n_bytes = std::min(in_p_stride, out_p_stride);
 
-  int x, y;
+  uint32_t x, y;
   for (y = 0; y < height; y++) {
     for (x = 0; x < n_bytes; x += 2) {
       out_p[y * out_p_stride + x + 0] = in_p[y * in_p_stride + x + 1];
